import(methods)
importFrom(igraph, plot.igraph, make_empty_graph, add_vertices, add_edges, topo_sort, permute)
importFrom(coda, effectiveSize, as.mcmc, as.mcmc.list)
importFrom(R6, R6Class)
importFrom(grDevices, dev.off, jpeg)
importFrom(graphics, lines, plot, text)
importFrom(stats, aggregate, ar, lm, optim, pnorm, qnorm, residuals, runif, sd, var, median, quantile, integrate, optimHess)
importFrom(utils, head)
importFrom(parallel, mclapply)
importFrom(pracma, hessian)
importFrom(numDeriv, genD, grad, jacobian)
useDynLib(nimble, .registration = TRUE)
S3method(as.matrix, CmodelValues)
S3method(as.matrix, modelValuesBaseClass)
S3method(as.list, CmodelValues)
S3method(as.list, modelValuesBaseClass)
S3method(length, nimPointerList)
export(calc_dmnormConjugacyContributions)
export(calc_dmnormAltParams)
export(calc_dwishAltParams)
export(calc_dcatConjugacyContributions)
export(CAR_calcM)
export(CAR_calcC)
export(CAR_calcCmatrix)
export(CAR_calcEVs2)
export(CAR_calcEVs3)
export(getNodeFunctionIndexedInfo)
export(singleModelValuesAccess)
export(getNimbleProject)
export(cc_getNodesInExpr)
export(calcAdaptationFactor)
export(is.nfGenerator)
export(mcmc_createModelObject)
export(nimbleInternalFunctions)
export(nimbleUserNamespace)
export(MCEM_mcse)
export(nimbleFunctionList)
export(nimbleType)
export("nfVar<-")
export("values<-")
export(ADNimbleList)
export(ADbreak)
export(ADproxyModelClass)
export(AGHQuad_params)
export(AGHQuad_summary)
export(CAR_calcNumIslands)
export(Rmatrix2mvOneVar)
export(any_na)
export(any_nan)
export(as.carAdjacency)
export(as.carCM)
export(asCol)
export(asRow)
export(autoBlock)
export(buildAGHQ)
export(buildAuxiliaryFilter)
export(buildBootstrapFilter)
export(buildEnsembleKF)
export(buildIteratedFilter2)
export(buildLaplace)
export(buildLiuWestFilter)
export(buildMCEM)
export(buildMCMC)
export(calcNodes)
export(calcNodesMV)
export(calculate)
export(calculateDiff)
export(calculateWAIC)
export(carBounds)
export(carMaxBound)
export(carMinBound)
export(cc_getNodesInExpr)
export(checkInterrupt)
export(clearCompiled)
export(cloglog)
export(compileNimble)
export(configureMCMC)
export(configureRJ)
export(cube)
export(dCRP)
export(dcar_normal)
export(dcar_proper)
export(dcat)
export(dconstraint)
export(ddexp)
export(ddirch)
export(decide)
export(decideAndJump)
export(declare)
export(deregisterDistributions)
export(dexp_nimble)
export(dflat)
export(dhalfflat)
export(dinterval)
export(dinvgamma)
export(dinvwish_chol)
export(dlkj_corr_cholesky)
export(dmnorm_chol)
export(dmulti)
export(dmvt_chol)
export(dsqrtinvgamma)
export(dt_nonstandard)
export(dwish_chol)
export(eigenNimbleList)
export(expit)
export(extractControlElement)
export(getBUGSexampleDir)
export(getBound)
export(getConditionallyIndependentSets)
export(getDefinition)
export(getDimension)
export(getDistributionInfo)
export(getLogProb)
export(getLogProbNodes)
export(getLogProbNodesMV)
export(getMacroParameters)
export(getNimbleOption)
export(getParam)
export(getParamNames)
export(getSamplesDPmeasure)
export(getType)
export(getsize)
export(icloglog)
export(identityMatrix)
export(ilogit)
export(initializeModel)
export(inprod)
export(inverse)
export(iprobit)
export(is.Cmodel)
export(is.Cnf)
export(is.Rmodel)
export(is.model)
export(is.nf)
export(is.nl)
export(isDiscrete)
export(isUserDefined)
export(logdet)
export(logfact)
export(loggam)
export(logit)
export(makeBoundInfo)
export(makeModelDerivsInfo)
export(makeParamInfo)
export(messageIfVerbose)
export(modelValues)
export(modelValuesConf)
export(model_macro_builder)
export(nfMethod)
export(nfVar)
export(nf_preProcessMemberDataObject)
export(nimArray)
export(nimC)
export(nimCat)
export(nimCopy)
export(nimDerivs)
export(nimDim)
export(nimEigen)
export(nimEquals)
export(nimInteger)
export(nimIntegrate)
export(nimLogical)
export(nimMatrix)
export(nimNumeric)
export(nimOptim)
export(nimOptimDefaultControl)
export(nimOptimMethod)
export(nimPrint)
export(nimRep)
export(nimRound)
export(nimSeq)
export(nimStep)
export(nimStop)
export(nimSvd)
export(nimSwitch)
export(nimbleCode)
export(nimbleExternalCall)
export(nimbleFunction)
export(nimbleFunctionVirtual)
export(nimbleList)
export(nimbleMCMC)
export(nimbleModel)
export(nimbleOptions)
export(nimbleRcall)
export(optimControlNimbleList)
export(optimDefaultControl)
export(optimResultNimbleList)
export(parameterTransform)
export(pdexp)
export(pexp_nimble)
export(phi)
export(pinvgamma)
export(pow)
export(pow_int)
export(pqDefined)
export(printErrors)
export(probit)
export(pt_nonstandard)
export(qdexp)
export(qexp_nimble)
export(qinvgamma)
export(qt_nonstandard)
export(rCRP)
export(rankSample)
export(rcar_normal)
export(rcar_proper)
export(rcat)
export(rconstraint)
export(rdexp)
export(rdirch)
export(readBUGSmodel)
export(registerDistributions)
export(resize)
export(rexp_nimble)
export(rflat)
export(rhalfflat)
export(rinterval)
export(rinvgamma)
export(rinvwish_chol)
export(rlkj_corr_cholesky)
export(rmnorm_chol)
export(rmulti)
export(rmvt_chol)
export(rsqrtinvgamma)
export(rt_nonstandard)
export(run.time)
export(runAGHQ)
export(runCrossValidate)
export(runLaplace)
export(runMCMC)
export(rwish_chol)
export(sampler_AF_slice)
export(sampler_BASE)
export(sampler_CAR_normal)
export(sampler_CAR_proper)
export(sampler_CRP)
export(sampler_CRP_concentration)
export(sampler_RJ_fixed_prior)
export(sampler_RJ_indicator)
export(sampler_RJ_toggled)
export(sampler_RW)
export(sampler_RW_block)
export(sampler_RW_block_lkj_corr_cholesky)
export(sampler_RW_dirichlet)
export(sampler_RW_lkj_corr_cholesky)
export(sampler_RW_llFunction)
export(sampler_RW_llFunction_block)
export(sampler_RW_multinomial)
export(sampler_RW_wishart)
export(sampler_binary)
export(sampler_categorical)
export(sampler_crossLevel)
export(sampler_ess)
export(sampler_noncentered)
export(sampler_polyagamma)
export(sampler_posterior_predictive)
export(sampler_prior_samples)
export(sampler_slice)
export(sampler_slice_CRP_base_param)
export(samplesSummary)
export(setAndCalculate)
export(setAndCalculateDiff)
export(setAndCalculateOne)
export(setSize)
export(setupMargNodes)
export(simNodes)
export(simNodesMV)
export(simulate)
export(stick_breaking)
export(summaryAGHQ)
export(summaryLaplace)
export(svdNimbleList)
export(testBUGSmodel)
export(valueInCompiledNimbleFunction)
export(values)
export(waicDetailsNimbleList)
export(waicNimbleList)
export(withNimbleOptions)
exportClasses(ADproxyModelClass)
exportClasses(CmodelBaseClass)
exportClasses(CnimbleFunctionBase)
exportClasses(MCMCconf)
exportClasses(RmodelBaseClass)
exportClasses(codeBlockClass)
exportClasses(modelBaseClass)
exportClasses(modelValuesBaseClass)
exportClasses(nimbleFunctionBase)
exportClasses(nimbleFunctionList)
exportClasses(nimbleType)
exportClasses(singleVarAccessClass)
