source("../../R/Rcheck.R")
d <- read.jagsdata("leuk-data.R")
inits <- read.jagsdata("leuk-init.R")
m <- jags.model("leuk.bug", d, inits,  n.chains=2)
update(m, 1000)
x <- coda.samples(m, "beta", n.iter=2000)
source("bench-test1.R")
check.fun()
