source("../../R/Rcheck.R")
load.module("glm")
data <- read.jagsdata("blocker-data.R")
inits <- read.jagsdata("blocker-init.R")
m <- jags.model("blocker.bug", data, inits, n.chains = 2)
update(m, 3000)
x <- coda.samples(m, c("d","delta.new","sigma"), n.iter=30000, thin=10)
source("bench-test1.R")
check.fun()
