source("../../R/Rcheck.R")
load.module("glm")
d <- read.jagsdata("salm-data.R")
inits <- read.jagsdata("salm-init.R")
m <- jags.model("salm-pareto.bug", d, inits, n.chains=2)
update(m, 1000)
load.module("dic")
x <- coda.samples(m, c("alpha","beta","gamma","sigma","deviance"), n.iter=5000)
source("bench-test2.R")
check.fun()
