source("../../R/Rcheck.R")
load.module("glm")
d <- read.jagsdata("lsat-data.R")
inits <- read.jagsdata("lsat2-init.R")
m <- jags.model("lsat2.bug", d, inits, n.chains=2, n.adapt=500)
update(m, 500)
x <- coda.samples(m, c("delta","eta"), thin=10, n.iter=10000)
source("bench-test2.R")
check.fun()
