model in mvotree.bug
data in mvotree-data.R
compile, nchains(2)
parameters in mvotree-inits.R
initialize
update 10000
monitor mu, thin(10)
monitor sigma, thin(10)
monitor sigmaC, thin(10)
update 100000
coda *

