load glm
model in schools.bug
data in schools-data.R
compile, nchains(2)
parameters in schools-inits.R
initialize
update 1000
monitor beta
monitor gamma
monitor phi
monitor theta
update 10000
coda *

