model in inhaler.bug
data in inhaler-data.R
compile, nchains(2)
parameters in inhaler-inits.R
initialize
update 1000
monitor a
monitor beta
monitor kappa
monitor log.sigma
monitor pi
monitor sigma
update 1000
coda *
exit
